local folder, core = ...

if not core.LibNameplate then
	return
end

local L = core.L or LibStub("AceLocale-3.0"):GetLocale(folder, true)
local _G = _G
local pairs_ = pairs
local GetTime_ = GetTime
local CreateFrame_ = CreateFrame
local table_remove = table.remove
local table_sort = table.sort
local type = type
local table_getn = table.getn
local Debug = core.Debug
local table_getn = table.getn
local DebuffTypeColor = DebuffTypeColor


-- Isolate the environment

--~ setfenv(1, core)
-- local

core.buffBars = {}


local P = {}
local buffFrames = core.buffFrames
local LibNameplate = core.LibNameplate
local nametoGUIDs = core.nametoGUIDs
local buffBars = core.buffBars
local buffFrames = core.buffFrames
local guidBuffs = core.guidBuffs


local prev_OnEnable = core.OnEnable;
function core:OnEnable()
	prev_OnEnable(self);
	
--~ 	Debug("OnEnable", "frames", self.db.profile)
	P = self.db.profile --this can change on profile change.
end

----------------------------------------------------------
local function UpdateBuffSize(frame, size)				--
-- Update a spell frame's texture size. 				--
-- This takes into adjusts height for cooldown size.	--
----------------------------------------------------------
	frame.texture:SetWidth(size)
	frame.texture:SetHeight(size)
	
	frame.border:SetHeight(size)
	frame.border:SetWidth(size)
	
	frame:SetWidth(size)
	
	if P.showCooldown == true then
		frame:SetHeight(size + frame.cd:GetStringHeight())
	else
		frame:SetHeight(size)
	end
end

------------------------------------------------------
local function UpdateBuffCDSize(buffFrame, size)	--
-- Set cooldown text size.							--
------------------------------------------------------
	buffFrame.cd:SetFont("Fonts\\FRIZQT__.TTF", size, "NORMAL")
	buffFrame.cdbg:SetHeight(buffFrame.cd:GetStringHeight())
--~ 	iconOnShow(buffFrame) --lazy way to hide/show cooldown.
end

--------------------------------------------------
local function SetStackSize(buffFrame, size)	--
-- Set the stack text size.						--
--------------------------------------------------
	buffFrame.stack:SetFont("Fonts\\FRIZQT__.TTF", size, "OUTLINE")
end

------------------------------------------
local function iconOnShow(self)			--
-- Called when spell frames are shown.	--
------------------------------------------
	self.alphaUp = false
	self:SetAlpha(1)
	if P.showCooldown == true and self.expirationTime > 0 then
		self.cdbg:Show()
		self.cd:Show()
	else
		self.cdbg:Hide()
		self.cd:Hide()
	end
	
	local iconSize = P.iconSize
	local cooldownSize = P.cooldownSize
	local stackSize = P.stackSize
	local spellName = self.spellName or "X"
	
	if P.spellOpts[spellName] then
		iconSize = P.spellOpts[spellName].iconSize or iconSize
		cooldownSize = P.spellOpts[spellName].cooldownSize or cooldownSize
		stackSize = P.spellOpts[spellName].stackSize or stackSize
	end

	if self.playerCast and P.biggerSelfSpells == true then
		UpdateBuffSize(self, iconSize * 1.2)
	else
		UpdateBuffSize(self, iconSize)
	end
	
	UpdateBuffCDSize(self, cooldownSize)

	if self.stackCount and self.stackCount > 1 then
		self.stack:Show()
		SetStackSize(self, stackSize)
	else
		self.stack:Hide()
	end
	
	if self.isDebuff then
		local colour = DebuffTypeColor[self.debuffType or ""];
		self.border:SetVertexColor(colour.r, colour.g, colour.b);
		self.border:Show()
	else
		self.border:Hide()
	end
end

------------------------------------------
local function iconOnHide(self)			--
-- Called when spell frames are shown.	--
------------------------------------------
	self.stack:Hide()
	UpdateBuffSize(self, P.iconSize)
end



----------------------------------------------
local function iconOnUpdate(self, elapsed)	--
-- Fires for spell frames.					--
----------------------------------------------
	self.lastUpdate = self.lastUpdate + elapsed
	if self.lastUpdate > 0.1 then --abit fast for cooldown flash.
		self.lastUpdate = 0

		if self.stackCount > 1 then
			self.stack:SetText(self.stackCount)
		end
		
		if self.expirationTime > 0 then
			local timeLeft = core:Round(self.expirationTime - GetTime_())
			if P.showCooldown == true then
				self.cd:SetText(core:SecondsToString(timeLeft, 1))
				self.cd:SetTextColor(core:RedToGreen(timeLeft, self.duration))
				self.cdbg:SetWidth(self.cd:GetStringWidth())
			end
			
			if (timeLeft / self.duration ) < .2 and timeLeft < 60 then --buff only has 20% timeleft and is less then 60 seconds.
				local alpha = self:GetAlpha()
				if alpha <= 0.2 then
					self.alphaUp = true
				elseif alpha >= 1 then
					self.alphaUp = false
				end
				if self.alphaUp == true then
					self:SetAlpha(alpha+0.2)
				else
					self:SetAlpha(alpha-0.2)
				end
			end
			
			if GetTime_() > self.expirationTime then
				self:Hide()
				
				local GUID = LibNameplate:GetGUID(self.realPlate)
				if GUID then
					core:AddBuffsToPlate(self.realPlate, GUID)
				else
					local plateName =  LibNameplate:GetName(self.realPlate)
					if plateName and nametoGUIDs[plateName] then
						core:AddBuffsToPlate(self.realPlate, nametoGUIDs[plateName])
					end
				end
			end
--~ 			else
--~ 				self.cd:Hide()
--~ 				self.cdbg:Hide()
		end
		
	end
end





local function SetBarSize(barFrame, width, height)
	barFrame:SetWidth(width)
	barFrame:SetHeight(height)
end



local function CreateBuffFrame(parentFrame, realPlate)
	local f = CreateFrame_("frame", nil, parentFrame) --
	f.realPlate = realPlate
	f:SetFrameStrata("LOW")
	
	f.texture = f:CreateTexture(nil,"BACKGROUND")
	f.texture:SetPoint("TOP", f)


	
	local cd = f:CreateFontString(nil, "ARTWORK", "ChatFontNormal")
--~ 	cd:SetFont("Fonts\\FRIZQT__.TTF", P.cooldownSize, "NORMAL") --done later.
	
	
	cd:SetText("0")
	cd:SetPoint("TOP", f.texture, "BOTTOM")
--~ 	cd:Show()
--~ 	cd:Hide()
	f.cd = cd

	--Make the text easier to see.
	f.cdbg = f:CreateTexture(nil,"LOW")
	f.cdbg:SetTexture(0,0,0,.75)
	f.cdbg:SetPoint("CENTER", cd)
--~ 	f.cdbg:SetHeight(cd:GetStringHeight()) --done later
--~ 	f.cdbg:Hide()
	
	f.stack = f:CreateFontString(nil, "ARTWORK", "ChatFontNormal")
--~ 	f.stack:SetFont("Fonts\\FRIZQT__.TTF", P.stackSize, "OUTLINE") --done later.

	f.stack:SetText("0")
	f.stack:SetPoint("BOTTOMRIGHT", f.texture, "BOTTOMRIGHT", -1, 3)
	f.stack:Hide()

	
	f.border = f:CreateTexture(nil,"OVERLAY")
	f.border:SetPoint("CENTER", f.texture)
	f.border:SetTexture("Interface\\Buttons\\UI-Debuff-Overlays")
	f.border:SetTexCoord(0.296875,0.5703125,0,0.515625);
--~ 	f.border:Show()
	
	
--~ 	UpdateBuffSize(f, P.iconSize)

--~ 	Debug("CreateBuffFrame", f:GetWidth(), f:GetHeight())
	
	f.lastUpdate = 0
	f.expirationTime = 0
	f:SetScript("OnShow", iconOnShow)
	f:SetScript("OnHide", iconOnHide)
	
	
	f:SetScript("OnUpdate", iconOnUpdate) 
	f.stackCount = 0
--~ 	f:SetScript("OnSizeChanged", function(self, ...)
--~ 		local bar = self:GetParent()
--~ 		UpdateBarSize(bar)
--~ 	end)

	return f
end

------------------------------------------
function core:UpdateBarsBackground()	--
-- Show/Hide bar background texture.	--
------------------------------------------
	for plate in pairs_(buffBars) do 
		for b in pairs_(buffBars[plate]) do 
		
			if P.showBarBackground == true then
				buffBars[plate][b].barBG:Show()
			else
				buffBars[plate][b].barBG:Hide()
			end
		
		end
	end
end
--~ local barSizes = {}

----------------------------------------------------------
local function CreatePlateBar(parentFrame, realPlate)	--
-- Create and return a bar frame.						--
----------------------------------------------------------
	local f = CreateFrame_("frame", nil, parentFrame) 
	f.realPlate = realPlate
	
	f:SetFrameStrata("BACKGROUND")
	f:SetFrameLevel(parentFrame:GetFrameLevel())
	
	f:SetWidth(1)
	f:SetHeight(1)
	
	--Make the text easier to see.
	f.barBG = f:CreateTexture(nil,"LOW")
	f.barBG:SetAllPoints(true)

	f.barBG:SetTexture(1,1,1,0.3) --uses for testing
	if P.showBarBackground == true then
		f.barBG:Show()
	else
		f.barBG:Hide()
	end
	
--~ 	f:Show()
	return f
end

----------------------------------------------------------------------
local function BuildPlateBars(plate, visibleFrame)					--
-- Build all our bar frames for a plate. 							--
-- We anchor these to the plate and our spell frames to the bar.	--
----------------------------------------------------------------------
	buffBars[plate] = buffBars[plate] or {}
	
	
	if not buffBars[plate][1] then
		buffBars[plate][1] = CreatePlateBar(visibleFrame, plate)
	end
	buffBars[plate][1]:ClearAllPoints()
	buffBars[plate][1]:SetPoint(P.barAnchorPoint, visibleFrame, P.plateAnchorPoint, P.barOffsetX, P.barOffsetY) -- + GetCooldownHeight(buffFrames[plate][1])
	
	
	if P.numRows > 1 then
		for r=2, P.numRows do
			if not buffBars[plate][r] then
				buffBars[plate][r] = CreatePlateBar(visibleFrame, plate)
			end
			buffBars[plate][r]:ClearAllPoints()
			
			
			if P.rowGroth == 1 then --up
				buffBars[plate][r]:SetPoint("BOTTOMLEFT", buffBars[plate][r-1], "TOPLEFT", 0, 0)
			else --down
				buffBars[plate][r]:SetPoint("TOPLEFT", buffBars[plate][r-1], "BOTTOMLEFT", 0, 0)
			end
		end
	end
end

--------------------------------------------------------------------------------------
local function UpdateBarSize(barFrame)												--
-- Update a bar's size taking into account all the spell frame's height and width.	--
--------------------------------------------------------------------------------------
	if barFrame:GetNumChildren() == 0 then
		Debug("Our buff bar lost it's children?")
		return
	end
	local totalWidth = 1
	local totalHeight = 1
	local c = {barFrame:GetChildren()}
	
	local frame
	for i=1, table_getn(c) do
		if i <= P.iconsPerRow then
			frame = c[i]
			totalWidth = totalWidth + frame:GetWidth()
			if frame:GetHeight() > totalHeight then
				totalHeight = frame:GetHeight()
			end
		end
	end
	
--~ 	Debug("UpdateBarSize", table_getn(c), P.iconsPerRow, totalHeight, totalWidth)
	barFrame:SetWidth(totalWidth)
	barFrame:SetHeight(totalHeight)
--~ 	SetBarSize(barFrame, totalWidth, totalHeight)
end

----------------------------------------------
function core:AddBuffsToPlate(plate, GUID)	--
-- Show spells on a plate linked to a GUID.	--
----------------------------------------------

	if not buffFrames[plate] or not buffFrames[plate][2]  then
		self:BuildBuffFrame(plate)
	end
	
	local t, f
--~ 	Debug("AddBuffsToPlate 0", GUID)
	if guidBuffs[GUID] then
	
		for i=(P.numRows * P.iconsPerRow), 1, -1 do 
			if guidBuffs[GUID][i] then
				if guidBuffs[GUID][i].expirationTime > 0 and GetTime_() > guidBuffs[GUID][i].expirationTime then
					table_remove(guidBuffs[GUID], i)
				end
			end
		end
	
		table_sort(guidBuffs[GUID], function(a,b) 
			if(a and b) then 
				return (a.playerCast or 0) > (b.playerCast or 0);
			end 
		end)

--~ 		Debug("AddBuffsToPlate 1", GUID)
		for i=1, P.numRows * P.iconsPerRow do 
--~ 			Debug("AddBuffsToPlate 2", GUID)
			if buffFrames[plate][i] then

				if guidBuffs[GUID][i] and (guidBuffs[GUID][i].expirationTime == 0 or guidBuffs[GUID][i].expirationTime > GetTime_()) then
					buffFrames[plate][i].spellName		= guidBuffs[GUID][i].name or ""
					buffFrames[plate][i].expirationTime	= guidBuffs[GUID][i].expirationTime or 0
					buffFrames[plate][i].duration		= guidBuffs[GUID][i].duration or 1
					buffFrames[plate][i].stackCount = guidBuffs[GUID][i].stackCount or 0
					buffFrames[plate][i].isDebuff = guidBuffs[GUID][i].isDebuff
					buffFrames[plate][i].debuffType = guidBuffs[GUID][i].debuffType
					buffFrames[plate][i].playerCast = guidBuffs[GUID][i].playerCast
					
					
					buffFrames[plate][i].texture:SetTexture(guidBuffs[GUID][i].icon)

					if buffFrames[plate][i]:IsShown() then
						iconOnShow(buffFrames[plate][i])
					else
						buffFrames[plate][i]:Show()--make sure OnShow fires.
					end
					iconOnUpdate(buffFrames[plate][i], 1) --do a onupdate.
				else
	--~ 				Debug("AddBuffsToPlate 4", i, GUID)
					buffFrames[plate][i]:Hide()
					buffFrames[plate][i].expirationTime = 0
				end
			end
		end
	end

	for r=P.numRows, 1, -1  do 
		UpdateBarSize(buffBars[plate][r])
	end
end

----------------------------------------------------------------------------------
function core:AddUnknownIcon(plate)												--
-- Display a question mark icon since we don't know the GUID of the plate/mob.	--
----------------------------------------------------------------------------------
	if not buffFrames[plate] then
		self:BuildBuffFrame(plate, nil, true)
	end

	local i = 1 --eaiser for me to copy/paste code elsewhere.
	buffFrames[plate][i].spellName		= false
	buffFrames[plate][i].expirationTime	= 0
	buffFrames[plate][i].duration		= 1
	buffFrames[plate][i].stackCount		= 0
	buffFrames[plate][i].isDebuff		= false
	buffFrames[plate][i].debuffType		= false
	buffFrames[plate][i].playerCast		= false
	
	
	buffFrames[plate][i].texture:SetTexture("Interface\\Icons\\Inv_misc_questionmark")
	
	if buffFrames[plate][i]:IsShown() then
		iconOnShow(buffFrames[plate][i])
	else
		buffFrames[plate][i]:Show()--make sure OnShow fires.
	end
	
--~ 	UpdateBarSize(buffBars[plate][1])
	for r=P.numRows, 1, -1  do 
		UpdateBarSize(buffBars[plate][r])
	end

end

----------------------------------------------------------
function core:ResetAllPlateIcons()						--
-- This will reset all the anchors on the spell frames.	--
----------------------------------------------------------
	for plate in pairs_(buffFrames) do 
		core:BuildBuffFrame(plate, true)
	end
end

------------------------------------------------------
function core:BuildBuffFrame(plate, reset, onlyOne)	--
-- Create our buff frames on a plate.				--
------------------------------------------------------

	local visibleFrame = LibNameplate:GetVisibleFrame(plate)
	if not buffBars[plate] then
		BuildPlateBars(plate, visibleFrame)
	end

	if not buffBars[plate][P.numRows] then--user increased the size.
		BuildPlateBars(plate, visibleFrame)
	end
	
	
	buffFrames[plate] = buffFrames[plate] or {}
	
	if reset then
		for i=1, table_getn(buffFrames[plate]) do 
			buffFrames[plate][i]:Hide() --makesure all frames stop OnUpdating.
		end
		
		for b in pairs_(buffBars[plate]) do 
			buffBars[plate][b]:Hide()
		end
	end
	
	local total = 1 --total number of spell frames
	if not buffFrames[plate][total] then
		buffFrames[plate][total] = CreateBuffFrame(buffBars[plate][1], plate)
	end
	buffFrames[plate][total]:SetParent(buffBars[plate][1])--Doing this seems to make UpdateBarSize() function properly. =/
	
	buffFrames[plate][total]:ClearAllPoints()
	buffFrames[plate][total]:SetPoint("TOPLEFT", buffBars[plate][1]) -- + GetCooldownHeight(buffFrames[plate][1])
	buffBars[plate][total]:Show()

	if onlyOne then
		--Save CPU cycles not creating every single frame. Lets wait till they're needed.
--~ 		UpdateBarSize(buffBars[plate][1])
		return
	end

	
	
	
	local prevFrame = buffFrames[plate][total]
	for i=2, P.iconsPerRow do 
		total = total + 1
		if not buffFrames[plate][total] then
			buffFrames[plate][total] = CreateBuffFrame(buffBars[plate][1], plate)
		end
		buffFrames[plate][total]:SetParent(buffBars[plate][1])

		buffFrames[plate][total]:ClearAllPoints()
		buffFrames[plate][total]:SetPoint("TOPLEFT", prevFrame, "TOPRIGHT")

		prevFrame = buffFrames[plate][total]
	end
--~ 	UpdateBarSize(buffBars[plate][1])

	
	if P.numRows > 1 then
		for r=2, P.numRows do 
			for i=1, P.iconsPerRow do 
				total = total + 1

				if not buffFrames[plate][total] then
					buffFrames[plate][total] = CreateBuffFrame(buffBars[plate][r], plate)
				end
				buffFrames[plate][total]:SetParent(buffBars[plate][r])
				
				
				buffFrames[plate][total]:ClearAllPoints()
				if i == 1 then
					buffFrames[plate][total]:SetPoint("TOPLEFT", buffBars[plate][r])
				else
					buffFrames[plate][total]:SetPoint("TOPLEFT", prevFrame, "TOPRIGHT")
				end
				
				prevFrame = buffFrames[plate][total]
			end

--~ 			UpdateBarSize(buffBars[plate][r])
			buffBars[plate][r]:Show()

		end
	end

end

--------------------------------------------------------------
function core:ResetBarPoint(barFrame, plate, visibleFrame)	--
-- Reset a bar's anchor point.								--
--------------------------------------------------------------

	barFrame:ClearAllPoints()
--~ 			buffFrames[plate][i]:SetPoint(P.barAnchorPoint, plate, P.plateAnchorPoint, P.barOffsetX, P.barOffsetY)
	local visibleFrame = visibleFrame or LibNameplate:GetVisibleFrame(plate)
	barFrame:SetParent(visibleFrame)
	barFrame:SetPoint(P.barAnchorPoint, visibleFrame, P.plateAnchorPoint, P.barOffsetX, P.barOffsetY) -- + GetCooldownHeight(barFrame)
end

--------------------------------------------------------------
function core:ResetIconSizes()								--
-- Reset all icon sizes. Called when user changes settings.	--
--------------------------------------------------------------

	local iconSize

	local frame
	for plate in pairs_(self.buffFrames) do 
		for i=1, table_getn(self.buffFrames[plate]) do 
			frame = self.buffFrames[plate][i]
			
			
			if P.spellOpts[frame.name] then
				iconSize = P.spellOpts[frame.name].iconSize
			else
				iconSize = P.iconSize
			end
			
			--Update the spell texture size.
			frame.texture:SetWidth(iconSize)
			frame.texture:SetHeight(iconSize)

			--Update the frame as a whole, this takes into account the size of the cooldown size.
			frame:SetWidth(iconSize)
			
			if P.showCooldown == true then
				frame:SetHeight(iconSize + frame.cd:GetStringHeight())
			else
				frame:SetHeight(iconSize)
			end
		end
	end
end

----------------------------------------------------------------------
function core:ResetCooldownSize()									--
-- Reset cooldown text sizes. Called when user changes settings.	--
----------------------------------------------------------------------
	for plate in pairs_(buffFrames) do 
		for i=1, table_getn(buffFrames[plate]) do 
--~ 			buffFrames[plate][i].cd:SetFontObject(core:GetCooldownFont())

			UpdateBuffCDSize(buffFrames[plate][i], buffFrames[plate][i].spellName and P.spellOpts[buffFrames[plate][i].spellName] and P.spellOpts[buffFrames[plate][i].spellName].cooldownSize or P.cooldownSize)
		end
	end
end



----------------------------------
function core:ResetStackSizes()	--
-- Update stack text size.		--
----------------------------------
	for plate in pairs_(buffFrames) do 
		for i=1, table_getn(buffFrames[plate]) do 
			SetStackSize(buffFrames[plate][i], buffFrames[plate][i].spellName and P.spellOpts[buffFrames[plate][i].spellName] and P.spellOpts[buffFrames[plate][i].spellName].stackSize or P.stackSize)
--~ 			buffFrames[plate][i].stack:SetFont("Fonts\\FRIZQT__.TTF", P.stackSize, "OUTLINE")
		end
	end
end

--------------------------------------
function core:ResetAllBarPoints()	--
-- Reset all bar anchors.			--
--------------------------------------
	for plate in pairs_(buffBars) do 
		if buffBars[plate][1] then
			self:ResetBarPoint(buffBars[plate][1], plate)
		end
		
		for r=2, table_getn(buffBars[plate]) do 
			buffBars[plate][r]:ClearAllPoints()
			if P.rowGroth == 1 then --up
				buffBars[plate][r]:SetPoint("BOTTOMLEFT", buffBars[plate][r-1], "TOPLEFT", 0, 0)
			else --down
				buffBars[plate][r]:SetPoint("TOPLEFT", buffBars[plate][r-1], "BOTTOMLEFT", 0, 0)
			end
		end
	end
end

------------------------------------------------------------------
function core:ShowAllKnownSpells()								--
-- When we change number of icons to show we hide all icons.	--
-- This will reshow the buffs again in their new locations.		--
------------------------------------------------------------------
	local GUID
	for plate in pairs_(buffFrames) do 
		local GUID = LibNameplate:GetGUID(plate)
		if GUID then
			self:AddBuffsToPlate(plate, GUID)
		else
			local plateName =  LibNameplate:GetName(plate)
			if plateName and nametoGUIDs[plateName] then
				self:AddBuffsToPlate(plate, nametoGUIDs[plateName])
			end
		end
	end
end